package httpclient

import (
	"bytes"
	"io"
	"io/ioutil"
	"net/http"
	"time"

	"github.com/gojektech/valkyrie"
	"github.com/pkg/errors"
)

// client is the http client implementation
type client struct {
	client Doer

	timeout    time.Duration
	retryCount int
	retrier    Retriable
}

const (
	defaultRetryCount  = 0
	defaultHTTPTimeout = 30 * time.Second
)

var _ Client = (*client)(nil)
var defaultClient *client

func init() {
	defaultClient = NewClient()
}

// NewClient returns a new instance of http client
func NewClient(opts ...Option) *client {
	client := client{
		timeout:    defaultHTTPTimeout,
		retryCount: defaultRetryCount,
		retrier:    NewNoRetrier(),
	}

	for _, opt := range opts {
		opt(&client)
	}

	if client.client == nil {
		client.client = &http.Client{
			Timeout: client.timeout,
		}
	}

	return &client
}

// Get makes a HTTP GET request to provided URL with default client
func Get(url string, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	return defaultClient.Get(url, headers, callbacks...)
}

// Get makes a HTTP GET request to provided URL
func (c *client) Get(url string, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	var response *http.Response
	request, err := http.NewRequest(http.MethodGet, url, nil)
	if err != nil {
		return response, errors.Wrap(err, "GET - request creation failed")
	}

	if headers != nil {
		request.Header = headers
	}
	for _, cb := range callbacks {
		cb(request)
	}

	return c.Do(request)
}

// Post makes a HTTP POST request to provided URL and requestBody with default client
func Post(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	return defaultClient.Post(url, body, headers, callbacks...)
}

// Post makes a HTTP POST request to provided URL and requestBody
func (c *client) Post(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	var response *http.Response
	request, err := http.NewRequest(http.MethodPost, url, body)
	if err != nil {
		return response, errors.Wrap(err, "POST - request creation failed")
	}

	if headers != nil {
		request.Header = headers
	}
	for _, cb := range callbacks {
		cb(request)
	}

	return c.Do(request)
}

// Put makes a HTTP PUT request to provided URL and requestBody with default client
func Put(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	return defaultClient.Put(url, body, headers, callbacks...)
}

// Put makes a HTTP PUT request to provided URL and requestBody
func (c *client) Put(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	var response *http.Response
	request, err := http.NewRequest(http.MethodPut, url, body)
	if err != nil {
		return response, errors.Wrap(err, "PUT - request creation failed")
	}

	if headers != nil {
		request.Header = headers
	}
	for _, cb := range callbacks {
		cb(request)
	}

	return c.Do(request)
}

// Patch makes a HTTP PATCH request to provided URL and requestBody with default client
func Patch(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	return defaultClient.Patch(url, body, headers, callbacks...)
}

// Patch makes a HTTP PATCH request to provided URL and requestBody
func (c *client) Patch(url string, body io.Reader, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	var response *http.Response
	request, err := http.NewRequest(http.MethodPatch, url, body)
	if err != nil {
		return response, errors.Wrap(err, "PATCH - request creation failed")
	}

	if headers != nil {
		request.Header = headers
	}
	for _, cb := range callbacks {
		cb(request)
	}

	return c.Do(request)
}

// Delete makes a HTTP DELETE request with provided URL with default client
func Delete(url string, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	return defaultClient.Delete(url, headers, callbacks...)
}

// Delete makes a HTTP DELETE request with provided URL
func (c *client) Delete(url string, headers http.Header, callbacks ...func(req *http.Request)) (*http.Response, error) {
	var response *http.Response
	request, err := http.NewRequest(http.MethodDelete, url, nil)
	if err != nil {
		return response, errors.Wrap(err, "DELETE - request creation failed")
	}

	if headers != nil {
		request.Header = headers
	}
	for _, cb := range callbacks {
		cb(request)
	}

	return c.Do(request)
}

// Do makes an HTTP request with the native `http.Do` interface
func (c *client) Do(request *http.Request) (*http.Response, error) {
	request.Close = true

	var bodyReader *bytes.Reader

	if request.Body != nil {
		reqData, err := ioutil.ReadAll(request.Body)
		if err != nil {
			return nil, err
		}
		bodyReader = bytes.NewReader(reqData)
		request.Body = ioutil.NopCloser(bodyReader) // prevents closing the body between retries
	}

	multiErr := &valkyrie.MultiError{}
	var response *http.Response

	for i := 0; i <= c.retryCount; i++ {
		if response != nil {
			response.Body.Close()
		}

		var err error
		response, err = c.client.Do(request)
		if bodyReader != nil {
			// Reset the body reader after the request since at this point it's already read
			// Note that it's safe to ignore the error here since the 0,0 position is always valid
			_, _ = bodyReader.Seek(0, 0)
		}

		if err != nil {
			multiErr.Push(err.Error())

			backoffTime := c.retrier.NextInterval(i)
			time.Sleep(backoffTime)
			continue
		}

		if response.StatusCode >= http.StatusInternalServerError {
			backoffTime := c.retrier.NextInterval(i)
			time.Sleep(backoffTime)
			continue
		}

		multiErr = &valkyrie.MultiError{} // Clear errors if any iteration succeeds
		break
	}

	return response, multiErr.HasError()
}
